{ Fast implementation of the Helix cipher.
  Helix is a high-speed stream cipher combined with MAC functionality.
  Implementation is made with mainly i386 assembly code.
  <P>It consumes around 7.2 clockcycles per byte encrypted / decrypted on a Athlon 1200 or
  around 10.2 clockcycles per byte encrypted / decrypted on a Pentium 4 2.8C for block sizes larger
  than 2kb and still fit into the processors cache. Note that Delphi does not allow to specify how
  to align individual procedures on specific 16 byte boundaries, so depending on the final layout of
  the executable / dll can the speed vary +-10%.
  <P>General usage:
  <PRE>
    procedure UseHelix;<BR>
    var<BR>
      H: THelix;<BR>
      Nonce,MAC: THelixBlock;<BR>
      B: array of Byte;<BR>
    const<BR>
      BlockSize = 16000;<BR>
      Key = 'Hello World';<BR>
    begin<BR>
      // And do some timing....<BR>
      HelixSetKey(H,Key[1],Length(Key));<BR>
      // Use an empty nonce (normally you will fill it with something useful).<BR>
      System.FillChar(Nonce,sizeof(Nonce),0);<BR>
      // Setup a buffer for the message to encrypt...<BR>
      SetLength(B,BlockSize);<BR>
      // Fill buffer here...<BR>
      HelixStart(H,Nonce);<BR>
      HelixMac(Helix,MAC,HelixEncrypt(H,B[0],BlockSize));<BR>
      // store MAC here...<BR>
      HelixBurn(H);<BR>
    end;<BR>
  </PRE>
  <P>Links: http://www.schneier.com/paper-helix.html or
            http://www.macfergus.com/helix/index.html . }
unit UnitHelix;

{ ContinuIT Free Library License (CFLL)

  This source code is owned and only owned by ContinuIT BV.
  (C) 2004 ContinuIT BV, All rights reserved.

  Permission is granted to anyone to use this software for any kind of application, and to
  alter it and redistribute it freely, subject to the following restrictions:
  - The origin of this software must not be misrepresented, you must not claim that you wrote
    the original software.
  - Altered source versions must be plainly marked as such, and must not be misrepresented as
    being the original software.
  - You may not create a library that uses this library as a main part and sell that library.
  - You must have a visible line in your programs about box and/or documentation stating that this
    ContinuIT library is used, with a link where this library can be found.
  - This notice may not be removed or altered from any source distribution.
  - This software is provided 'as-is', without any expressed or implied warranty. In no event will
    the author(s) or ContinuIT BV be held liable for any damages arising from the use of this
    software.

  This license is subject to Dutch Law. }

{ 20040120  ritsaert@continuit.nl     Initial version.
  20040122  ritsaert@continuit.nl     Added HelixEncryptNoWrite and HelixDecryptNoWrite.
  20040123  ritsaert@continuit.nl     Added HelixEncryptCopy and HelixDecryptCopy. }

interface

type
  { The state of a Helix stream cipher is kept within this structure.
    <P>This is a record instead of a class to avoid RTTI class info being created. }
  THelix = record
    { Expanded key. This is filled by @link(HelixSetKey). }
    K: array [0..7] of Longword;
    { Expanded partial Key X1. Initialized by @link(HelixStart). }
    X1: array [0..7] of Longword;
    { Working registers. Initialized by @link(HelixStart) and updated while encrypting, decrypting,
      calculating MAC etc. }
    Z: array [0..4] of Longword;      // Z_0..Z_4   Working registers
    { Number of blocks processed within this stream plus 8. }
    h_iplus8: Longword;               // Block number (only lower 32 bits)
    { Length of the unexpanded key times 4. This is filled by @link(HelixSetKey). }
    ULen_times4: Longword;
    { Length of the unexpanded key plus 64. This is filled by @link(HelixSetKey). }
    ULen_plus64: Longword;
    { Number of blocks to operate on for the encrypting / decrypting asm routines.
      Will be zero on exit. }
    BlockCount: Integer;
    { Pointer to the first block to operate on for the encrypting / decrypting asm routines.
      Pointer will be advanced after operation. }
    BlockSource: PLongword;
    { Destination pointer used only by @link(HelixEncryptCopy) and @link(HelixDecryptCopy).
      and their used assembler routines. }
    BlockDest: PLongword;
  end;

  { A MAC and Nonce are 16 bytes large and are represented by this datatype. }
  THelixBlock = array [0..3] of Longword;

{ This routine expands a given key into a Helix structure. It Fills K[], ULen_times4, ULen_plus64.
  The unexpanded key should be 0..32 bytes long.
  <P>Returns: boolean denoting success. The routine is always succesful if a valid key is given. }
function HelixSetKey(var Helix: THelix; const Key; KeyLength: Integer): Boolean;

{ This routine expands the nonce and resets the Helix structure to the initial state.
  After this call @link(HelixEncrypt) and @link(HelixDecrypt) can be called.
  It fills X1[], Z[], h_iplus8 to initial state. @link(HelixSetKey) must have been called. }
procedure HelixStart(var Helix: THelix; const Nonce: THelixBlock);

{ Encrypt a block of data using the current Helix state.
  <P>You can call this procedure repeatedly as long as the non-last call all have a DataLength
  of modulo 4 equals 0. Before you can call this function, call @link(HelixStart).
  <P>If you are interested in a MAC, call @link(HelixMac) after this function.
  <P>Returns: DataLength mod 4. This value can be used in the call to @link(HelixMac). }
function HelixEncrypt(var Helix: THelix; var Data; DataLength: Integer): Integer;

{ Decrypt a block of data using the current Helix state.
  <P>You can call this procedure repeatedly as long as the non-last call all have a DataLength
  of modulo 4 equals 0. Before you can call this function, call @link(HelixStart).
  <P>If you are interested in a MAC, call @link(HelixMac) after this function.
  <P>Returns: DataLength mod 4. This value can be used in the call to @link(HelixMac). }
function HelixDecrypt(var Helix: THelix; var Data; DataLength: Integer): Integer;

{ Calculates the MAC of a Helix message. This procedure is called after
  calls to @link(HelixEncrypt) or @link(HelixDecrypt) or @link(HelixStart) if you want the
  MAC of an empty message. The parameter n passed is the Length of the message modulo 4. }
procedure HelixMac(var Helix: THelix; out Mac: THelixBlock; n: Integer);

{ Cleanup of possible sensitive data contained within Helix structure.
  This invalidates all data within the structure including the expanded key. }
procedure HelixBurn(var Helix: THelix);

{ Encrypt a block of data using the current Helix state but the ciphertext produced is NOT saved.
  Although no ciphertext is produced, the Helix state is updated so @link(HelixMac) will produce
  the correct result.
  <P>You can call this procedure repeatedly as long as the non-last call all have a DataLength
  of modulo 4 equals 0. Before you can call this function, call @link(HelixStart).
  <P>Returns: DataLength mod 4. This value can be used in the call to @link(HelixMac). }
function HelixEncryptNoWrite(var Helix: THelix; var Data; DataLength: Integer): Integer;

{ Decrypt a block of data using the current Helix state but the plaintext produced is NOT saved.
  Although no plaintext is produced, the Helix state is updated so @link(HelixMac) will produce
  the correct result.
  <P>You can call this procedure repeatedly as long as the non-last call all have a DataLength
  of modulo 4 equals 0. Before you can call this function, call @link(HelixStart).
  <P>Returns: DataLength mod 4. This value can be used in the call to @link(HelixMac). }
function HelixDecryptNoWrite(var Helix: THelix; var Data; DataLength: Integer): Integer;

{ Encrypt a block of data using the current Helix state. The ciphertext produced is written in
  a different buffer, leaving the plaintext intact.
  <P>You can call this procedure repeatedly as long as the non-last call all have a DataLength
  of modulo 4 equals 0. Before you can call this function, call @link(HelixStart).
  <P>If you are interested in a MAC, call @link(HelixMac) after this function.
  <P>Returns: DataLength mod 4. This value can be used in the call to @link(HelixMac). }
function HelixEncryptCopy(var Helix: THelix; const SourceData;
                          var DestData; DataLength: Integer): Integer;

{ Decrypt a block of data using the current Helix state. The plaintext produced is written in
  a different buffer, leaving the ciphertext intact.
  <P>You can call this procedure repeatedly as long as the non-last call all have a DataLength
  of modulo 4 equals 0. Before you can call this function, call @link(HelixStart).
  <P>If you are interested in a MAC, call @link(HelixMac) after this function.
  <P>Returns: DataLength mod 4. This value can be used in the call to @link(HelixMac). }
function HelixDecryptCopy(var Helix: THelix; const SourceData;
                          var DestData; DataLength: Integer): Integer;

implementation

// HELIX BLOCK ROUTINES WRITTEN IN ASSEMBLER FOR SPEED.

{ processes one block without writing data in or out; only the helix status is updated. }
procedure HelixBlock(var Helix: THelix; P_i: Longword); register;
asm
  // Save registers we're not allowed to change due to calling convention
  push ebp; push ebx; push esi; push edi;
  mov ecx,[eax + THelix.Z + 0];
  mov ebp,[eax + THelix.Z + 4];
  mov ebx,[eax + THelix.Z + 8];
  mov esi,[eax + THelix.Z + 12];
  mov edi,[eax + THelix.Z + 16];
  push edx; // We need it for some calculations
  // Note : eax = @Helix; P_i: pushed on stack
  // AddRol(A,D,15); AddRol(B,E,25); XorRol(C,A,9); XorRol(D,B,10); AddRol(E,C,17);
  add ecx,esi; rol esi,15;
  add ebp,edi; ror edi,32 - 25;
  xor ebx,ecx; rol ecx,9;
  xor esi,ebp; rol ebp,10;
  add edi,ebx; ror ebx,32 - 17;
  // A ^= D + K[i]; D := Rol(D,30); XorRol(B,E,13); AddRol(C,A,20); AddRol(D,B,11); XorRol(E,C,5);
  mov edx,[eax + THelix.H_iplus8]; and edx,7; mov edx,[eax + THelix.K + 4 * edx]; // K[i]
  add edx,esi; xor ecx, edx; ror esi,32 - 30;
  xor ebp,edi; rol edi, 13;
  add ebx,ecx; ror ecx, 32 - 20;
  add esi,ebp; rol ebp, 11;
  xor edi,ebx; rol ebx, 5;
  // A += D ^ P_i; D := Rol(D,15); AddRol(B,E,25); XorRol(C,A,9); XorRol(D,B,10); AddRol(E,C,17);
  pop edx; xor edx, esi; add ecx,edx; rol esi,15;
  add ebp,edi; ror edi,32 - 25;
  xor ebx,ecx; rol ecx,9;
  xor esi,ebp; rol ebp,10;
  add edi,ebx; ror ebx,32 - 17;
  // A ^= D + X1[i] + h_iplus8++; D := Rol(D,30); XorRol(B,E,13); AddRol(C,A,20); AddRol(D,B,11); XorRol(E,C,5);
  mov edx,[eax + THelix.H_iplus8]; and edx,7; mov edx,[eax + THelix.X1 + 4 * edx]; // X1[i]
  add edx,[eax + THelix.H_iplus8]; add edx,esi; xor ecx,edx; ror esi,32 - 30;
  inc dword ptr [eax + THelix.H_iplus8];
  xor ebp,edi; rol edi, 13;
  add ebx,ecx; ror ecx, 32 - 20;
  add esi,ebp; rol ebp, 11;
  xor edi,ebx; rol ebx, 5;
  // Save mixed result.
  mov [eax + THelix.Z + 0], ecx;
  mov [eax + THelix.Z + 4], ebp;
  mov [eax + THelix.Z + 8], ebx;
  mov [eax + THelix.Z + 12], esi;
  mov [eax + THelix.Z + 16], edi;
  // Restore registers
  pop edi; pop esi; pop ebx; pop ebp;
end;

{ Encrypt a given amount of blocks and store the ciphertext in place. }
procedure HelixBlockEnc(var Helix: THelix); register;
asm
  // Save registers we're not allowed to change due to calling convention
  push ebp; push ebx; push esi; push edi;
  mov ecx,[eax + THelix.Z + 0];
  mov ebp,[eax + THelix.Z + 4];
  mov ebx,[eax + THelix.Z + 8];
  mov esi,[eax + THelix.Z + 12];
  mov edi,[eax + THelix.Z + 16];
  // Note : eax = @Helix; P_i: pushed on stack each time
@loop:
  mov edx,[eax + THelix.BlockSource];                    // edx ::= P
  push dword ptr [edx];                                  // P_i := P^;
  xor  [edx],ecx;                                        // P^ := W xor Helix.Z[0];
  lea edx,[edx + 4]; mov [eax + THelix.BlockSource],edx; // Inc(P);
  // AddRol(A,D,15); AddRol(B,E,25); XorRol(C,A,9); XorRol(D,B,10); AddRol(E,C,17);
  add ecx,esi; rol esi,15;
  add ebp,edi; ror edi,32 - 25;
  xor ebx,ecx; rol ecx,9;
  xor esi,ebp; rol ebp,10;
  add edi,ebx; ror ebx,32 - 17;
  // A ^= D + K[i]; D := Rol(D,30); XorRol(B,E,13); AddRol(C,A,20); AddRol(D,B,11); XorRol(E,C,5);
  mov edx,[eax + THelix.H_iplus8]; and edx,7; mov edx,[eax + THelix.K + 4 * edx]; // K[i]
  add edx,esi; xor ecx, edx; ror esi,32 - 30;
  xor ebp,edi; rol edi, 13;
  add ebx,ecx; ror ecx, 32 - 20;
  add esi,ebp; rol ebp, 11;
  xor edi,ebx; rol ebx, 5;
  // A += D ^ P_i; D := Rol(D,15); AddRol(B,E,25); XorRol(C,A,9); XorRol(D,B,10); AddRol(E,C,17);
  pop edx; xor edx, esi; add ecx,edx; rol esi,15;
  add ebp,edi; ror edi,32 - 25;
  xor ebx,ecx; rol ecx,9;
  xor esi,ebp; rol ebp,10;
  add edi,ebx; ror ebx,32 - 17;
  // A ^= D + X1[i] + h_iplus8++; D := Rol(D,30); XorRol(B,E,13); AddRol(C,A,20); AddRol(D,B,11); XorRol(E,C,5);
  mov edx,[eax + THelix.H_iplus8]; and edx,7; mov edx,[eax + THelix.X1 + 4 * edx]; // X1[i]
  add edx,[eax + THelix.H_iplus8]; add edx,esi; xor ecx,edx; ror esi,32 - 30;
  inc dword ptr [eax + THelix.H_iplus8];
  xor ebp,edi; rol edi, 13;
  add ebx,ecx; ror ecx, 32 - 20;
  add esi,ebp; rol ebp, 11;
  xor edi,ebx; rol ebx, 5;
  // loop?
  dec dword ptr [eax + THelix.BlockCount];
  jnz @loop;
  // Save mixed result.
  mov [eax + THelix.Z + 0], ecx;
  mov [eax + THelix.Z + 4], ebp;
  mov [eax + THelix.Z + 8], ebx;
  mov [eax + THelix.Z + 12], esi;
  mov [eax + THelix.Z + 16], edi;
  // Restore registers
  pop edi; pop esi; pop ebx; pop ebp;
end;

{ Encrypt the given amount of blocks. Only update the Heli state, no ciphertext is stored. }
procedure HelixBlockEncNoWrite(var Helix: THelix); register;
asm
  // Save registers we're not allowed to change due to calling convention
  push ebp; push ebx; push esi; push edi;
  mov ecx,[eax + THelix.Z + 0];
  mov ebp,[eax + THelix.Z + 4];
  mov ebx,[eax + THelix.Z + 8];
  mov esi,[eax + THelix.Z + 12];
  mov edi,[eax + THelix.Z + 16];
  // Note : eax = @Helix; P_i: pushed on stack each time
@loop:
  mov edx,[eax + THelix.BlockSource];                    // edx ::= P
  push dword ptr [edx];                                  // P_i := P^;
  lea edx,[edx + 4]; mov [eax + THelix.BlockSource],edx; // Inc(P);
  // AddRol(A,D,15); AddRol(B,E,25); XorRol(C,A,9); XorRol(D,B,10); AddRol(E,C,17);
  add ecx,esi; rol esi,15;
  add ebp,edi; ror edi,32 - 25;
  xor ebx,ecx; rol ecx,9;
  xor esi,ebp; rol ebp,10;
  add edi,ebx; ror ebx,32 - 17;
  // A ^= D + K[i]; D := Rol(D,30); XorRol(B,E,13); AddRol(C,A,20); AddRol(D,B,11); XorRol(E,C,5);
  mov edx,[eax + THelix.H_iplus8]; and edx,7; mov edx,[eax + THelix.K + 4 * edx]; // K[i]
  add edx,esi; xor ecx, edx; ror esi,32 - 30;
  xor ebp,edi; rol edi, 13;
  add ebx,ecx; ror ecx, 32 - 20;
  add esi,ebp; rol ebp, 11;
  xor edi,ebx; rol ebx, 5;
  // A += D ^ P_i; D := Rol(D,15); AddRol(B,E,25); XorRol(C,A,9); XorRol(D,B,10); AddRol(E,C,17);
  pop edx; xor edx, esi; add ecx,edx; rol esi,15;
  add ebp,edi; ror edi,32 - 25;
  xor ebx,ecx; rol ecx,9;
  xor esi,ebp; rol ebp,10;
  add edi,ebx; ror ebx,32 - 17;
  // A ^= D + X1[i] + h_iplus8++; D := Rol(D,30); XorRol(B,E,13); AddRol(C,A,20); AddRol(D,B,11); XorRol(E,C,5);
  mov edx,[eax + THelix.H_iplus8]; and edx,7; mov edx,[eax + THelix.X1 + 4 * edx]; // X1[i]
  add edx,[eax + THelix.H_iplus8]; add edx,esi; xor ecx,edx; ror esi,32 - 30;
  inc dword ptr [eax + THelix.H_iplus8];
  xor ebp,edi; rol edi, 13;
  add ebx,ecx; ror ecx, 32 - 20;
  add esi,ebp; rol ebp, 11;
  xor edi,ebx; rol ebx, 5;
  // loop?
  dec dword ptr [eax + THelix.BlockCount];
  jnz @loop;
  // Save mixed result.
  mov [eax + THelix.Z + 0], ecx;
  mov [eax + THelix.Z + 4], ebp;
  mov [eax + THelix.Z + 8], ebx;
  mov [eax + THelix.Z + 12], esi;
  mov [eax + THelix.Z + 16], edi;
  // Restore registers
  pop edi; pop esi; pop ebx; pop ebp;
end;

{ Encrypt a given amount of blocks and store the ciphertext in another buffer. }
procedure HelixBlockEncCopy(var Helix: THelix); register;
asm
  // Save registers we're not allowed to change due to calling convention
  push ebp; push ebx; push esi; push edi;
  mov ecx,[eax + THelix.Z + 0];
  mov ebp,[eax + THelix.Z + 4];
  mov ebx,[eax + THelix.Z + 8];
  mov esi,[eax + THelix.Z + 12];
  mov edi,[eax + THelix.Z + 16];
  // Note : eax = @Helix; P_i: pushed on stack each time
@loop:
  push ebx;                                               // Save Z[2]; we need another register :-(
  mov  ebx,[eax + THelix.BlockSource];                    // ebx ::= P
  mov  edx,[ebx]                                          // edx ::= P_i
  lea  ebx,[ebx + 4]; mov [eax + THelix.BlockSource],ebx; // Inc(P);
  mov  ebx,[eax + THelix.BlockDest];                      // ebx ::= Q
  mov  [ebx],edx; xor [ebx],ecx                           // Q^ ::= P_i xor Z[0];
  lea  ebx,[ebx + 4]; mov [eax + THelix.BlockDest],ebx;   // Inc(Q);
  pop  ebx; push edx;                                     // Restore Z[2], Push P_i
  // AddRol(A,D,15); AddRol(B,E,25); XorRol(C,A,9); XorRol(D,B,10); AddRol(E,C,17);
  add ecx,esi; rol esi,15;
  add ebp,edi; ror edi,32 - 25;
  xor ebx,ecx; rol ecx,9;
  xor esi,ebp; rol ebp,10;
  add edi,ebx; ror ebx,32 - 17;
  // A ^= D + K[i]; D := Rol(D,30); XorRol(B,E,13); AddRol(C,A,20); AddRol(D,B,11); XorRol(E,C,5);
  mov edx,[eax + THelix.H_iplus8]; and edx,7; mov edx,[eax + THelix.K + 4 * edx]; // K[i]
  add edx,esi; xor ecx, edx; ror esi,32 - 30;
  xor ebp,edi; rol edi, 13;
  add ebx,ecx; ror ecx, 32 - 20;
  add esi,ebp; rol ebp, 11;
  xor edi,ebx; rol ebx, 5;
  // A += D ^ P_i; D := Rol(D,15); AddRol(B,E,25); XorRol(C,A,9); XorRol(D,B,10); AddRol(E,C,17);
  pop edx; xor edx, esi; add ecx,edx; rol esi,15;
  add ebp,edi; ror edi,32 - 25;
  xor ebx,ecx; rol ecx,9;
  xor esi,ebp; rol ebp,10;
  add edi,ebx; ror ebx,32 - 17;
  // A ^= D + X1[i] + h_iplus8++; D := Rol(D,30); XorRol(B,E,13); AddRol(C,A,20); AddRol(D,B,11); XorRol(E,C,5);
  mov edx,[eax + THelix.H_iplus8]; and edx,7; mov edx,[eax + THelix.X1 + 4 * edx]; // X1[i]
  add edx,[eax + THelix.H_iplus8]; add edx,esi; xor ecx,edx; ror esi,32 - 30;
  inc dword ptr [eax + THelix.H_iplus8];
  xor ebp,edi; rol edi, 13;
  add ebx,ecx; ror ecx, 32 - 20;
  add esi,ebp; rol ebp, 11;
  xor edi,ebx; rol ebx, 5;
  // loop?
  dec dword ptr [eax + THelix.BlockCount];
  jnz @loop;
  // Save mixed result.
  mov [eax + THelix.Z + 0], ecx;
  mov [eax + THelix.Z + 4], ebp;
  mov [eax + THelix.Z + 8], ebx;
  mov [eax + THelix.Z + 12], esi;
  mov [eax + THelix.Z + 16], edi;
  // Restore registers
  pop edi; pop esi; pop ebx; pop ebp;
end;

{ Decrypt a given amount of blocks and store the plaintext in place. }
procedure HelixBlockDec(var Helix: THelix); register;
asm
  // Save registers we're not allowed to change due to calling convention
  push ebp; push ebx; push esi; push edi;
  mov ecx,[eax + THelix.Z + 0];
  mov ebp,[eax + THelix.Z + 4];
  mov ebx,[eax + THelix.Z + 8];
  mov esi,[eax + THelix.Z + 12];
  mov edi,[eax + THelix.Z + 16];
  // Note : eax = @Helix; P_i: pushed on stack each time
@loop:
  mov  edx,[eax + THelix.BlockSource];                   // edx ::= P
  xor  [edx],ecx;                                        // P^ := P^ xor Helix.Z[0];
  push dword ptr [edx];                                  // P_i := P^;
  lea edx,[edx + 4]; mov [eax + THelix.BlockSource],edx; // Inc(P);
  // AddRol(A,D,15); AddRol(B,E,25); XorRol(C,A,9); XorRol(D,B,10); AddRol(E,C,17);
  add ecx,esi; rol esi,15;
  add ebp,edi; ror edi,32 - 25;
  xor ebx,ecx; rol ecx,9;
  xor esi,ebp; rol ebp,10;
  add edi,ebx; ror ebx,32 - 17;
  // A ^= D + K[i]; D := Rol(D,30); XorRol(B,E,13); AddRol(C,A,20); AddRol(D,B,11); XorRol(E,C,5);
  mov edx,[eax + THelix.H_iplus8]; and edx,7; mov edx,[eax + THelix.K + 4 * edx]; // K[i]
  add edx,esi; xor ecx, edx; ror esi,32 - 30;
  xor ebp,edi; rol edi, 13;
  add ebx,ecx; ror ecx, 32 - 20;
  add esi,ebp; rol ebp, 11;
  xor edi,ebx; rol ebx, 5;
  // A += D ^ P_i; D := Rol(D,15); AddRol(B,E,25); XorRol(C,A,9); XorRol(D,B,10); AddRol(E,C,17);
  pop edx; xor edx, esi; add ecx,edx; rol esi,15;
  add ebp,edi; ror edi,32 - 25;
  xor ebx,ecx; rol ecx,9;
  xor esi,ebp; rol ebp,10;
  add edi,ebx; ror ebx,32 - 17;
  // A ^= D + X1[i] + h_iplus8++; D := Rol(D,30); XorRol(B,E,13); AddRol(C,A,20); AddRol(D,B,11); XorRol(E,C,5);
  mov edx,[eax + THelix.H_iplus8]; and edx,7; mov edx,[eax + THelix.X1 + 4 * edx]; // X1[i]
  add edx,[eax + THelix.H_iplus8]; add edx,esi; xor ecx,edx; ror esi,32 - 30;
  inc dword ptr [eax + THelix.H_iplus8];
  xor ebp,edi; rol edi, 13;
  add ebx,ecx; ror ecx, 32 - 20;
  add esi,ebp; rol ebp, 11;
  xor edi,ebx; rol ebx, 5;
  // loop?
  dec dword ptr [eax + THelix.BlockCount];
  jnz @loop;
  // Save mixed result.
  mov [eax + THelix.Z + 0], ecx;
  mov [eax + THelix.Z + 4], ebp;
  mov [eax + THelix.Z + 8], ebx;
  mov [eax + THelix.Z + 12], esi;
  mov [eax + THelix.Z + 16], edi;
  // Restore registers
  pop edi; pop esi; pop ebx; pop ebp;
end;

{ Decrypt a given amount of blocks. The plaintext is not stored, only the Helix state is updated. }
procedure HelixBlockDecNoWrite(var Helix: THelix); register;
asm
  // Save registers we're not allowed to change due to calling convention
  push ebp; push ebx; push esi; push edi;
  mov ecx,[eax + THelix.Z + 0];
  mov ebp,[eax + THelix.Z + 4];
  mov ebx,[eax + THelix.Z + 8];
  mov esi,[eax + THelix.Z + 12];
  mov edi,[eax + THelix.Z + 16];
  // Note : eax = @Helix; P_i: pushed on stack each time
@loop:
  mov  edx,[eax + THelix.BlockSource];                   // edx ::= P
  lea edx,[edx + 4]; mov [eax + THelix.BlockSource],edx; // Inc(P);
  mov edx,[edx - 4]; xor edx,ecx; push edx;              // P_i := P^ xor Helix.Z[0];
  // AddRol(A,D,15); AddRol(B,E,25); XorRol(C,A,9); XorRol(D,B,10); AddRol(E,C,17);
  add ecx,esi; rol esi,15;
  add ebp,edi; ror edi,32 - 25;
  xor ebx,ecx; rol ecx,9;
  xor esi,ebp; rol ebp,10;
  add edi,ebx; ror ebx,32 - 17;
  // A ^= D + K[i]; D := Rol(D,30); XorRol(B,E,13); AddRol(C,A,20); AddRol(D,B,11); XorRol(E,C,5);
  mov edx,[eax + THelix.H_iplus8]; and edx,7; mov edx,[eax + THelix.K + 4 * edx]; // K[i]
  add edx,esi; xor ecx, edx; ror esi,32 - 30;
  xor ebp,edi; rol edi, 13;
  add ebx,ecx; ror ecx, 32 - 20;
  add esi,ebp; rol ebp, 11;
  xor edi,ebx; rol ebx, 5;
  // A += D ^ P_i; D := Rol(D,15); AddRol(B,E,25); XorRol(C,A,9); XorRol(D,B,10); AddRol(E,C,17);
  pop edx; xor edx, esi; add ecx,edx; rol esi,15;
  add ebp,edi; ror edi,32 - 25;
  xor ebx,ecx; rol ecx,9;
  xor esi,ebp; rol ebp,10;
  add edi,ebx; ror ebx,32 - 17;
  // A ^= D + X1[i] + h_iplus8++; D := Rol(D,30); XorRol(B,E,13); AddRol(C,A,20); AddRol(D,B,11); XorRol(E,C,5);
  mov edx,[eax + THelix.H_iplus8]; and edx,7; mov edx,[eax + THelix.X1 + 4 * edx]; // X1[i]
  add edx,[eax + THelix.H_iplus8]; add edx,esi; xor ecx,edx; ror esi,32 - 30;
  inc dword ptr [eax + THelix.H_iplus8];
  xor ebp,edi; rol edi, 13;
  add ebx,ecx; ror ecx, 32 - 20;
  add esi,ebp; rol ebp, 11;
  xor edi,ebx; rol ebx, 5;
  // loop?
  dec dword ptr [eax + THelix.BlockCount];
  jnz @loop;
  // Save mixed result.
  mov [eax + THelix.Z + 0], ecx;
  mov [eax + THelix.Z + 4], ebp;
  mov [eax + THelix.Z + 8], ebx;
  mov [eax + THelix.Z + 12], esi;
  mov [eax + THelix.Z + 16], edi;
  // Restore registers
  pop edi; pop esi; pop ebx; pop ebp;
end;

{ Decrypt a given amount of blocks and store the plaintext in another buffer. }
procedure HelixBlockDecCopy(var Helix: THelix); register;
asm
  // Save registers we're not allowed to change due to calling convention
  push ebp; push ebx; push esi; push edi;
  mov ecx,[eax + THelix.Z + 0];
  mov ebp,[eax + THelix.Z + 4];
  mov ebx,[eax + THelix.Z + 8];
  mov esi,[eax + THelix.Z + 12];
  mov edi,[eax + THelix.Z + 16];
  // Note : eax = @Helix; P_i: pushed on stack each time
@loop:
  push ecx                                                // Save Z[0]; we need another register :-(
  mov  edx,[eax + THelix.BlockSource];                    // edx ::= P
  xor  ecx,[edx];                                         // P_i := P^ xor Z[0];
  lea  edx,[edx + 4]; mov [eax + THelix.BlockSource],edx; // Inc(P);
  mov  edx,ecx                                            // move over
  mov  ecx,[eax + THelix.BlockDest];                      // ecx ::= Q
  mov  [ecx],edx                                          // Q^ := P_i;
  lea  ecx,[ecx + 4]; mov [eax + THelix.BlockDest],ecx;   // Inc(Q);
  pop  ecx; push edx;                                     // Restore Z[0], save P_i on top of stack
  // AddRol(A,D,15); AddRol(B,E,25); XorRol(C,A,9); XorRol(D,B,10); AddRol(E,C,17);
  add ecx,esi; rol esi,15;
  add ebp,edi; ror edi,32 - 25;
  xor ebx,ecx; rol ecx,9;
  xor esi,ebp; rol ebp,10;
  add edi,ebx; ror ebx,32 - 17;
  // A ^= D + K[i]; D := Rol(D,30); XorRol(B,E,13); AddRol(C,A,20); AddRol(D,B,11); XorRol(E,C,5);
  mov edx,[eax + THelix.H_iplus8]; and edx,7; mov edx,[eax + THelix.K + 4 * edx]; // K[i]
  add edx,esi; xor ecx, edx; ror esi,32 - 30;
  xor ebp,edi; rol edi, 13;
  add ebx,ecx; ror ecx, 32 - 20;
  add esi,ebp; rol ebp, 11;
  xor edi,ebx; rol ebx, 5;
  // A += D ^ P_i; D := Rol(D,15); AddRol(B,E,25); XorRol(C,A,9); XorRol(D,B,10); AddRol(E,C,17);
  pop edx; xor edx, esi; add ecx,edx; rol esi,15;
  add ebp,edi; ror edi,32 - 25;
  xor ebx,ecx; rol ecx,9;
  xor esi,ebp; rol ebp,10;
  add edi,ebx; ror ebx,32 - 17;
  // A ^= D + X1[i] + h_iplus8++; D := Rol(D,30); XorRol(B,E,13); AddRol(C,A,20); AddRol(D,B,11); XorRol(E,C,5);
  mov edx,[eax + THelix.H_iplus8]; and edx,7; mov edx,[eax + THelix.X1 + 4 * edx]; // X1[i]
  add edx,[eax + THelix.H_iplus8]; add edx,esi; xor ecx,edx; ror esi,32 - 30;
  inc dword ptr [eax + THelix.H_iplus8];
  xor ebp,edi; rol edi, 13;
  add ebx,ecx; ror ecx, 32 - 20;
  add esi,ebp; rol ebp, 11;
  xor edi,ebx; rol ebx, 5;
  // loop?
  dec dword ptr [eax + THelix.BlockCount];
  jnz @loop;
  // Save mixed result.
  mov [eax + THelix.Z + 0], ecx;
  mov [eax + THelix.Z + 4], ebp;
  mov [eax + THelix.Z + 8], ebx;
  mov [eax + THelix.Z + 12], esi;
  mov [eax + THelix.Z + 16], edi;
  // Restore registers
  pop edi; pop esi; pop ebx; pop ebp;
end;

procedure HelixKeyMix(var Helix: THelix); register;
asm
  push esi; push edi; push ebp; push ebx;
  // Loop 4 times..
  mov edx,4
@loop:
  //    A = K[0]; B = K[1]; C = K[2]; D = K[3]; E = U_len + 64;
  mov ecx,[eax + THelix.K];
  mov ebx,[eax + THelix.K + 4];
  mov ebp,[eax + THelix.K + 8];
  mov edi,[eax + THelix.K + 12];
  mov esi,[eax + THelix.ULen_plus64];

  add ecx,edi; rol edi,15;            // A += D; D <<<= 15;
  add ebx,esi; ror esi,(32-25);       // B += E; E <<<= 25;
  xor ebp,ecx; rol ecx,9;             // C ^= A; A <<<=  9;
  xor edi,ebx; rol ebx,10;            // D ^= B; B <<<= 10;
  add esi,ebp; ror ebp,(32-17);       // E += C; C <<<= 17;

  xor ecx,edi; ror edi,(32-30);       // A ^= D; D <<<= 30;
  xor ebx,esi; rol esi,13;            // B ^= E; E <<<= 13;
  add ebp,ecx; ror ecx,(32-20);       // C += A; A <<<= 20;
  add edi,ebx; rol ebx,11;            // D += B; B <<<= 11;
  xor esi,ebp; rol ebp,5;             // E ^= C; C <<<=  5;

  add ecx,edi; rol edi,15;            // A += D; D <<<= 15;
  add ebx,esi; ror esi,(32-25);       // B += E; E <<<= 25;
  xor ebp,ecx; rol ecx,9;             // C ^= A; A <<<=  9;
  xor edi,ebx; rol ebx,10;            // D ^= B; B <<<= 10;
  add esi,ebp; ror ebp,(32-17);       // E += C; C <<<= 17;

  xor ecx,edi; ror edi,(32-30);       // A ^= D; D <<<= 30;
  xor ebx,esi; {rol esi,13;}          // B ^= E; E <<<= 13;
  add ebp,ecx; ror ecx,(32-20);       // C += A; A <<<= 20;
  add edi,ebx; rol ebx,11;            // D += B; B <<<= 11;
  {xor esi,ebp;} rol ebp,5;           // E ^= C; C <<<=  5;

  // K[4] ^= A; K[5] ^= B; K[6] ^= C; K[7] ^= D;
  //  A = K[4]; B = K[5]; C = K[6]; D = K[7]; E = U_len + 64;
  xor ecx,[eax + THelix.K + 16];
  xor ebx,[eax + THelix.K + 20];
  xor ebp,[eax + THelix.K + 24];
  xor edi,[eax + THelix.K + 28];
  mov [eax + THelix.K + 16],ecx;
  mov [eax + THelix.K + 20],ebx;
  mov [eax + THelix.K + 24],ebp;
  mov [eax + THelix.K + 28],edi;
  mov esi,[eax + THelix.ULen_plus64];

  add ecx,edi; rol edi,15;            // A += D; D <<<= 15;
  add ebx,esi; ror esi,(32-25);       // B += E; E <<<= 25;
  xor ebp,ecx; rol ecx,9;             // C ^= A; A <<<=  9;
  xor edi,ebx; rol ebx,10;            // D ^= B; B <<<= 10;
  add esi,ebp; ror ebp,(32-17);       // E += C; C <<<= 17;

  xor ecx,edi; ror edi,(32-30);       // A ^= D; D <<<= 30;
  xor ebx,esi; rol esi,13;            // B ^= E; E <<<= 13;
  add ebp,ecx; ror ecx,(32-20);       // C += A; A <<<= 20;
  add edi,ebx; rol ebx,11;            // D += B; B <<<= 11;
  xor esi,ebp; rol ebp,5;             // E ^= C; C <<<=  5;

  add ecx,edi; rol edi,15;            // A += D; D <<<= 15;
  add ebx,esi; ror esi,(32-25);       // B += E; E <<<= 25;
  xor ebp,ecx; rol ecx,9;             // C ^= A; A <<<=  9;
  xor edi,ebx; rol ebx,10;            // D ^= B; B <<<= 10;
  add esi,ebp; ror ebp,(32-17);       // E += C; C <<<= 17;

  xor ecx,edi; ror edi,(32-30);       // A ^= D; D <<<= 30;
  xor ebx,esi; {rol esi,13;}          // B ^= E; E <<<= 13;
  add ebp,ecx; ror ecx,(32-20);       // C += A; A <<<= 20;
  add edi,ebx; rol ebx,11;            // D += B; B <<<= 11;
  {xor esi,ebp;} rol ebp,5;           // E ^= C; C <<<=  5;

  // K[0] ^= A; K[1] ^= B; K[2] ^= C; K[3] ^= D;
  xor [eax + THelix.K], ecx;
  xor [eax + THelix.K + 4], ebx;
  xor [eax + THelix.K + 8], ebp;
  xor [eax + THelix.K + 12], edi;
  // and... loop
  dec edx; jnz @loop
  // Cleanup
  pop ebx; pop ebp; pop edi; pop esi;
end;


// Fills K[], ULen_times4, ULen_plus64.
function HelixSetKey(var Helix: THelix; const Key; KeyLength: Integer): Boolean;
begin
  // Sanity check
  if (Keylength < 0) or (KeyLength > sizeof(Helix.K)) then begin
    Result := False;
    Exit;
  end;

  // 3.7 Key schedule.
  // Save key length
  Helix.ULen_times4 := Longword(KeyLength) shl 2;
  Helix.ULen_plus64 := Longword(KeyLength) + 64;

  // Copy and zero extend the key bytes
  System.FillChar(Helix.K, sizeof(Helix.K), 0);
  System.Move(Key, Helix.K, KeyLength);

  // And perform the key scedule step.
  HelixKeyMix(Helix);

  // We're succesful.
  Result := True;
end;

// HelixSetKey must have been called.
// Fills X1, Z, h_iplus8 to initial state.
procedure HelixStart(var Helix: THelix; const Nonce: THelixBlock);
begin
  with Helix do begin
    // 3.3 nonce setup
    // note: we assume (i + 8) shr 31 = 0. This is correct for messages < 8GB - 32B.
    X1[0] := K[4] + Nonce[0];
    X1[1] := K[5] + Nonce[1] + ULen_times4;
    X1[2] := K[6] + Nonce[2];
    X1[3] := K[7] + Nonce[3];
    X1[4] := K[0] - Nonce[0];
    X1[5] := K[1] - Nonce[1] + ULen_times4 + 1;
    X1[6] := K[2] - Nonce[2] + 2;
    X1[7] := K[3] - Nonce[3] + 3;

    // 3.4 initialisation
    h_iplus8 := 0;  // i = -8.
    Z[0] := K[3] xor Nonce[0];
    Z[1] := K[4] xor Nonce[1];
    Z[2] := K[5] xor Nonce[2];
    Z[3] := K[6] xor Nonce[3];
    Z[4] := K[7];

    // mixing..
    HelixBlock(Helix,0); HelixBlock(Helix,0);
    HelixBlock(Helix,0); HelixBlock(Helix,0);
    HelixBlock(Helix,0); HelixBlock(Helix,0);
    HelixBlock(Helix,0); HelixBlock(Helix,0);
  end;
end;

function FetchBytes(P: Pointer; Count: Integer): Longword; register;
asm
  // Assume Count 1..3
  mov ecx,eax;                 // ecx ::= P
  xor eax,eax                  // Result := 0;
  cmp edx,2
  jge @large
  mov al,byte ptr[ecx]         // Count = 1 --> Fetch a single byte and we're done.
  ret
@large:                        // Count = 2 or 3
  mov ax,word ptr [ecx];       // Fetch 2 bytes
  je @exit;                    // if Count was 2, we're done.
  ror eax,8
  mov ah,byte ptr [ecx + 2];
  rol eax,8                    // fetch the last byte
@exit:
end;

procedure StoreBytes(P: Pointer; Bytes: Longword; Count: Integer); register;
asm
  // eax ::= P, edx ::= Bytes, ecx ::= Count
  cmp ecx,2
  jge @large
  mov byte ptr [eax], dl;      // Count = 1 --> Store a single byte and we're done.
  ret
@large:                        // Count = 2 or 3
  mov word ptr [eax], dx;      // Store 2 bytes
  je @exit;                    // If count was 2, we're done
  shr edx,8;                   // Shift last byte into dh
  mov byte ptr [eax + 2], dh;  // and store...
@exit:
end;

{ You can call this function multiple times as long as the non-last call all get DataLength
  mod 4 = 0. }
function HelixEncrypt(var Helix: THelix; var Data; DataLength: Integer): Integer;
var
  P: PLongword;
  X: Longword;
begin
  // 3.5 encryption
  P := @Data;
  // Full blocks
  if DataLength >= 4 then begin
    Helix.BlockCount := DataLength shr 2;
    Helix.BlockSource := P;
    HelixBlockEnc(Helix);
    P := Helix.BlockSource;
    DataLength := DataLength and 3;
  end;
  // Last block
  if DataLength <> 0 then begin
    X := FetchBytes(P,DataLength);
    StoreBytes(P,X xor Helix.Z[0],DataLength);
    HelixBlock(Helix, X);
  end;
  // Return DataLength mod 4 (which it is right now) for possible MAC calculation.
  Result := DataLength;
end;

function HelixEncryptNoWrite(var Helix: THelix; var Data; DataLength: Integer): Integer;
var
  P: PLongword;
  X: Longword;
begin
  // 3.5 encryption
  P := @Data;
  // Full blocks
  if DataLength >= 4 then begin
    Helix.BlockCount := DataLength shr 2;
    Helix.BlockSource := P;
    HelixBlockEncNoWrite(Helix);
    P := Helix.BlockSource;
    DataLength := DataLength and 3;
  end;
  // Last block
  if DataLength <> 0 then HelixBlock(Helix, FetchBytes(P,DataLength));
  // Return DataLength mod 4 (which it is right now) for possible MAC calculation.
  Result := DataLength;
end;

// todo: HelixEncryptCopy
function HelixEncryptCopy(var Helix: THelix; const SourceData;
                          var DestData; DataLength: Integer): Integer;
var
  P, Q: PLongword;
  X: Longword;
begin // todo: stayed here
  // 3.5 encryption

  P := @SourceData;
  Q := @DestData;
  // Full blocks
  if DataLength >= 4 then begin
    Helix.BlockCount := DataLength shr 2;
    Helix.BlockSource := P;
    Helix.BlockDest := Q;
    HelixBlockEncCopy(Helix);
    P := Helix.BlockSource;
    Q := Helix.BlockDest;
    DataLength := DataLength and 3;
  end;
  // Last block
  if DataLength <> 0 then begin
    X := FetchBytes(P,DataLength);
    StoreBytes(Q,X xor Helix.Z[0],DataLength);
    HelixBlock(Helix, X);
  end;
  // Return DataLength mod 4 (which it is right now) for possible MAC calculation.
  Result := DataLength;
end;

function HelixDecrypt(var Helix: THelix; var Data; DataLength: Integer): Integer;
const
  AndMask: array [0..3] of Longword = ( $00000000, $000000FF, $0000FFFF, $00FFFFFF );
var
  P: PLongword;
  X: Longword;
begin
  // 3.6 decryption
  P := @Data;
  // Full blocks
  if DataLength >= 4 then begin
    Helix.BlockCount := DataLength shr 2;
    Helix.BlockSource := P;
    HelixBlockDec(Helix);
    P := Helix.BlockSource;
    DataLength := DataLength and 3;
  end;
  // Last block
  if DataLength <> 0 then begin
    X := (FetchBytes(P,DataLength) xor Helix.Z[0]) and AndMask[DataLength];
    StoreBytes(P,X,DAtaLength);
    HelixBlock(Helix, X);
  end;
  // Return DataLength mod 4 (which it is right now) for possible MAC calculation.
  Result := DataLength;
end;

function HelixDecryptNoWrite(var Helix: THelix; var Data; DataLength: Integer): Integer;
const
  AndMask: array [0..3] of Longword = ( $00000000, $000000FF, $0000FFFF, $00FFFFFF );
var
  P: PLongword;
  X: Longword;
begin
  // 3.6 decryption
  P := @Data;
  // Full blocks
  if DataLength >= 4 then begin
    Helix.BlockCount := DataLength shr 2;
    Helix.BlockSource := P;
    HelixBlockDecNoWrite(Helix);
    P := Helix.BlockSource;
    DataLength := DataLength and 3;
  end;
  // Last block
  if DataLength <> 0 then begin
    HelixBlock(Helix, (FetchBytes(P,DataLength) xor Helix.Z[0]) and AndMask[DataLength]);
  end;
  // Return DataLength mod 4 (which it is right now) for possible MAC calculation.
  Result := DataLength;
end;

// todo: HelixDecryptCopy
function HelixDecryptCopy(var Helix: THelix; const SourceData;
                          var DestData; DataLength: Integer): Integer;
const
  AndMask: array [0..3] of Longword = ( $00000000, $000000FF, $0000FFFF, $00FFFFFF );
var
  P, Q: PLongword;
  X: Longword;
begin
  // 3.6 decryption
  P := @SourceData;
  Q := @DestData;
  // Full blocks
  if DataLength >= 4 then begin
    Helix.BlockCount := DataLength shr 2;
    Helix.BlockSource := P;
    Helix.BlockDest := Q;
    HelixBlockDecCopy(Helix);
    P := Helix.BlockSource;
    Q := Helix.BlockDest;
    DataLength := DataLength and 3;
  end;
  // Last block
  if DataLength <> 0 then begin
    X := (FetchBytes(P,DataLength) xor Helix.Z[0]) and AndMask[DataLength];
    StoreBytes(Q,X,DAtaLength);
    HelixBlock(Helix, X);
  end;
  // Return DataLength mod 4 (which it is right now) for possible MAC calculation.
  Result := DataLength;
end;


// Calculates the MAC of a Helix message.
// Needs to be called after encryption or decryption
// n = MessageLength mod 4.
procedure HelixMac(var Helix: THelix; out Mac: THelixBlock; n: Integer);
begin
  // 3.6 compute MAC
  with Helix do begin
    Z[0] := Z[0] xor $912d94f1;
    HelixBlock(Helix,n);
    HelixBlock(Helix,n);
    HelixBlock(Helix,n);
    HelixBlock(Helix,n);
    HelixBlock(Helix,n);
    HelixBlock(Helix,n);
    HelixBlock(Helix,n);
    HelixBlock(Helix,n);
    Mac[0] := Z[0]; HelixBlock(Helix, n);
    Mac[1] := Z[0]; HelixBlock(Helix, n);
    Mac[2] := Z[0]; HelixBlock(Helix, n);
    Mac[3] := Z[0];
  end;
end;

procedure HelixBurn(var Helix: THelix);
begin
  System.FillChar(Helix,sizeof(Helix),0);
end;

end.
